/* src/config.h.  Generated from config.h.in by configure.  */
/* config.h.in.  Generated from configure.in by autoheader.  */

/* Define to 1 if you have the <getopt.h> header file. */
#define HAVE_GETOPT_H 1

#define HAVE_IOCTL 1

/* Define if you have the <libintl.h> header file.  */
#define HAVE_LIBINTL_H 1

/* Define if you have the <locale.h> header file.  */
#define HAVE_LOCALE_H 1

/* Define this if you want the NLS support.  */
#define ENABLE_NLS 1

/* Define to 1 if you have the <gnugetopt/getopt.h> header file. */
/* #undef HAVE_GNUGETOPT_GETOPT_H */

/* Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/* Define to 1 if you have the <memory.h> header file. */
#define HAVE_MEMORY_H 1

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT ""

/* Define to the full name of this package. */
#define PACKAGE_NAME "wput"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "wput 0.6"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "wput"

/* Define to the version of this package. */
#define PACKAGE_VERSION "0.6"

/* Define to 1 if you have the ANSI C header files. */
#define STDC_HEADERS 1

/* Define if all libs needed for ssl support are existing */
#define HAVE_SSL 1

/* Define to 1 if you have the long long type */
#define HAVE_LONG_LONG 1

/* Define to 1 if you have the <sys/termio.h> header file. */
/* #undef HAVE_SYSTERMIO */

/* Define to 1 if you have the <sys/winsize.h> header file. */
/* #undef HAVE_WINSIZE */

/* Define to 1 if you have the <termio.h> header file. */
#define HAVE_TERMIO 1


#define SIZEOF_SHORT 2
#define SIZEOF_INT 4
